from abc import abstractmethod
from typing import Callable, List
from uuid import uuid4

from Akatosh.universe import Mundus

from .event import InstantEvent
from .resource import Resource
from .states import State
from .logger import logger


class Entity:
    def __init__(
        self,
        label: str | None = None,
        create_at: int | float | Callable | None = None,
        terminate_at: int | float | Callable | None = None,
    ) -> None:
        """Create a entity.

        Args:
            label (str | None, optional): short description of the entity. Defaults to None.
            create_at (int | float | Callable | None, optional): when the life cycle of this entity should start. Defaults to None, then must call create() method manually.
            terminate_at (int | float | Callable | None, optional): when the life cycle of this entity should end. Defaults to None, then must call terminate() method manually.
        """
        self._id = uuid4().int
        self._label = label
        self._state: List[str] = list()
        self._occupied_resources: List[Resource] = list()
        self._creation: InstantEvent = None  # type: ignore
        self._termination: InstantEvent = None  # type: ignore

        if create_at is not None:
            if callable(create_at):
                self.create(at=round(create_at(), Mundus.resolution))
            else:
                self.create(at=round(create_at, Mundus.resolution))

        if terminate_at is not None:
            if callable(terminate_at):
                self.terminate(at=round(terminate_at(), Mundus.resolution))
            else:
                self.terminate(at=round(terminate_at, Mundus.resolution))

    def create(self, at: int | float) -> None:
        """The creation of the entity."""
        def _create():
            if self.terminated:
                raise RuntimeError(f"Entity {self.label} is already terminated.")
            self._state.append(State.CREATED)
            self.on_creation()
            logger.debug(f"Entity {self.label} created at {Mundus.now}")

        self._creation = InstantEvent(
            at=at, action=_create, label=f"Creation of {self.label}"
        )

    @abstractmethod
    def on_creation(self):
        """Callback function upon creation of the entity"""
        pass

    def terminate(self, at: int | float) -> None:
        """The termination of the entity."""
        def _terminate():
            if not self.created:
                raise RuntimeError(f"Entity {self.label} is not created yet.")
            self._state.append(State.TERMINATED)
            self.release_resources()
            self.on_termination()
            logger.debug(f"Entity {self.label} terminated at {Mundus.now}")

        self._termination = InstantEvent(
            at=at, action=_terminate, label=f"Termination of {self.label}"
        )

    @abstractmethod
    def on_termination(self):
        """Callback function upon termination of the entity"""
        pass

    def get(
        self, resource: Resource, amount: int | float | Callable | None = None
    ) -> None:
        """Consume certain amout from a resource. if no amount is specified, consume all."""
        if amount:
            if callable(amount):
                resource.distribute(self, amount())
                self.ocupied_resources.append(resource)
            else:
                resource.distribute(self, amount)
                self.ocupied_resources.append(resource)
        else:
            resource.distribute(self, resource.amount)

    def put(
        self, resource: Resource, amount: int | float | Callable | None = None
    ) -> None:
        """Return certain amout to a resource. if no amount is specified, return all."""
        if amount:
            if callable(amount):
                resource.collect(self, amount())
                if self not in resource.users:
                    self.ocupied_resources.remove(resource)
            else:
                resource.put(amount)
                if self not in resource.users:
                    self.ocupied_resources.remove(resource)
        else:
            resource.put(resource.amount)

    def release_resources(self):
        """Release all occupied resources."""
        for res in self.ocupied_resources:
            res.collect(self)

    @property
    def label(self):
        """Return the label of the entity."""
        return self._label

    @property
    def state(self):
        """Return the state(s) of the entity."""
        return self._state

    @property
    def created(self):
        """Return True if the entity is created."""
        return State.CREATED in self.state

    @property
    def terminated(self):
        """Return True if the entity is terminated."""
        return State.TERMINATED in self.state

    @property
    def ocupied_resources(self):
        """Return the resources occupied by the entity."""
        return self._occupied_resources



from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import Any, Callable, List
from uuid import uuid4

from .logger import logger
from .states import State
from .universe import Mundus


class Event:
    def __init__(
        self,
        at: int | float | Callable,
        precursor: Event | List[Event] | None = None,
        action: Callable | None = None,
        priority: int | float | Callable = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Base class for all events.

        Args:
            at (int | float | Callable): the time when the event happens.
            precursor (Event | List[Event] | None, optional): the precursor events. Defaults to None.
            action (Callable | None, optional): the actual action of the event, must be defined. Defaults to None.
            priority (int | float | Callable, optional): the priority of the event. When multiple event happen at the same time, the event with lower priority value will happen first. Defaults to 0.
            label (str | None, optional): Short descirption of the event. Defaults to None.
        """
        self._id = uuid4().int
        # set the time
        if callable(at):
            self._at = round(at(), Mundus.resolution)
        else:
            self._at = round(at, Mundus.resolution)
        # set the target event(s) that this event should wait and follow
        if isinstance(precursor, list):
            self._precursor = precursor
        else:
            self._precursor = [precursor] if precursor is not None else []
        self._follower: List[Event] = []
        for event in self.precursor:
            event._follower.append(self)
        # set the state
        if precursor:
            self.state = State.INACTIVE
        else:
            self.state = State.ACTIVE
        # set the action and the priority
        if action:
            self._action = action
        if callable(priority):
            self._priority = priority()
        else:
            self._priority = priority
        # assign label
        self._label = label
        # add to the universe
        Mundus.future_events.append(self)

    def end(self):
        """End the event and activate the follower events if there is any."""
        self.state = State.ENDED
        for event in self.follower:
            try:
                event.activate()
            except RuntimeError:
                logger.debug(f"Event {event.label} passed due time.")

    def activate(self, force: bool = False):
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")

        if force:
            self.state = State.ACTIVE
            logger.debug(f"Event {self.label} is set to active.")
        else:
            if all(e.state == State.ENDED for e in self.precursor):
                self.state = State.ACTIVE
                logger.debug(f"Event {self.label} is set to active.")
            else:
                raise RuntimeError(
                    f"Event {self.label} is waiting for other events to end."
                )

    def deactivate(self):
        """Deactivate the event.

        Raises:
            RuntimeError: raise if the event has already ended.
        """
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")
        self.state = State.INACTIVE

    def cancel(self):
        """Cancel the event.

        Raises:
            RuntimeError: raise if the event has already ended.
        """
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")
        self.state = State.CANCELED

    @abstractmethod
    async def _perform(self):
        """Abstract method for the event to perform its action."""
        pass

    def __eq__(self, _o: Event) -> bool:
        """Determine if two events are the same."""
        return self.id == _o.id

    @property
    def id(self) -> int:
        """Return the unique id of the event."""
        return self._id

    @property
    def label(self) -> str | None:
        """Return the label of the event."""
        return self._label

    @property
    def at(self) -> int | float:
        """Return the time when the event happens."""
        return self._at

    @property
    def precursor(self) -> List[Event]:
        """Return the precursor events."""
        return self._precursor

    @property
    def follower(self) -> List[Event]:
        """Return the follower events."""
        return self._follower

    @property
    def action(self) -> Callable | None:
        """Return the action of the event."""
        return self._action

    @property
    def priority(self) -> int | float:
        """Return the priority of the event."""
        return self._priority

    @property
    def state(self) -> str:
        """Return the state of the event."""
        return self._state

    @state.setter
    def state(self, state: str):
        """Set the state of the event."""
        self._state = state

    @property
    def active(self) -> bool:
        """Return True if the event is active."""
        return State.ACTIVE == self.state

    @property
    def inactive(self) -> bool:
        """Return True if the event is inactive."""
        return State.INACTIVE == self.state

    @property
    def cancelled(self) -> bool:
        """Return True if the event is cancelled."""
        return State.CANCELED == self.state

    @property
    def ended(self) -> bool:
        """Return True if the event is ended."""
        return State.ENDED == self.state


class InstantEvent(Event):
    """Instant event is an event that happens at a specific time and only happens once."""

    def __init__(
        self,
        at: int | float | Callable[..., Any],
        precursor: Event | List[Event] | None = None,
        action: Callable[..., Any] | None = None,
        priority: int | float | Callable[..., Any] = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Create a instant event.

        Args:
            at (int | float | Callable): the time when the event happens.
            precursor (Event | List[Event] | None, optional): the precursor events. Defaults to None.
            action (Callable | None, optional): the actual action of the event, must be defined. Defaults to None.
            priority (int | float | Callable, optional): the priority of the event. When multiple event happen at the same time, the event with lower priority value will happen first. Defaults to 0.
            label (str | None, optional): Short descirption of the event. Defaults to None.
        """
        super().__init__(at, precursor, action, priority, label, **kwargs)

    async def _perform(self):
        """Perform the action of the event."""
        if self.state == State.ACTIVE:
            if self.action:
                if inspect.iscoroutinefunction(self.action):
                    await self.action()
                else:
                    self.action()
            Mundus.current_events.remove(self)
            Mundus.past_events.append(self)
            self.end()
            logger.debug(f"Event {self.label} is ended.")


def event(
    at: int | float | Callable,
    precursor: Event | List[Event] | None = None,
    priority: int | float | Callable = 0,
    label: str | None = None,
    **kwargs,
):
    """Decorator for creating instant event.

    Args:
        at (int | float | Callable): When the event happens.
        precursor (Event | List[Event] | None, optional): The precursor events. Defaults to None.
        priority (int | float | Callable, optional): The priority of the event. Defaults to 0.
        label (str | None, optional): Short description of the event. Defaults to None.
    """

    def _event(func: Callable):
        return InstantEvent(
            at=at,
            precursor=precursor,
            action=func,
            priority=priority,
            label=label,
            **kwargs,
        )

    return _event


class ContinuousEvent(Event):
    def __init__(
        self,
        at: int | float | Callable[..., Any],
        interval: int | float | Callable[..., Any],
        duration: int | float | Callable[..., Any],
        precursor: Event | List[Event] | None = None,
        action: Callable[..., Any] | None = None,
        priority: int | float | Callable[..., Any] = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Continous event is an event that happens at a specific time and happens multiple times for a specific duration.

        Args:
            at (int | float | Callable[..., Any]): when the event happens.
            interval (int | float | Callable[..., Any]): the interval between each action repeat.
            duration (int | float | Callable[..., Any]): the duration of the event.
            precursor (Event | List[Event] | None, optional): the precursor events, can be both instant events or continous events. Defaults to None.
            action (Callable[..., Any] | None, optional): the actual action of the event. Defaults to None.
            priority (int | float | Callable[..., Any], optional): the priority of the event. Defaults to 0.
            label (str | None, optional): short description of the event.. Defaults to None.
        """
        super().__init__(
            at=at,
            precursor=precursor,
            action=action,
            priority=priority,
            label=label,
            **kwargs,
        )
        self._interval = interval
        if callable(duration):
            self._duration = round(duration(), Mundus.resolution)
        else:
            self._duration = round(duration, Mundus.resolution)
        self._till = self.at + self.duration
        self._sub_events: List[InstantEvent] = list()

    async def _perform(self):
        """Perform the action of the event. Compare to instant event, continous event will perform the action multiple times."""
        if self.action:
            self.action()
            if callable(self.interval):
                self._at = round(self.interval() + Mundus.now, Mundus.resolution)
            else:
                self._at = round(self.interval + Mundus.now, Mundus.resolution)
            if self.at <= self.till:
                logger.debug(f"Event {self.label} next step is at {self.at}.")
                Mundus.future_events.append(self)
            else:
                self.end()
                logger.debug(f"Event {self.label} is ended.")
            Mundus.current_events.remove(self)
            Mundus.past_events.append(self)

    @property
    def interval(self) -> int | float | Callable[..., Any]:
        """Return the interval between each action repeat."""
        return self._interval

    @property
    def duration(self) -> int | float:
        """Return the duration of the event."""
        return self._duration

    @property
    def till(self) -> int | float:
        """Return the time when the event ends."""
        return self._till

    @property
    def sub_events(self) -> List[InstantEvent]:
        """Return the sub events of the continous event."""
        return self._sub_events



from __future__ import annotations

from typing import Callable, List, Tuple

from .logger import logger
from .universe import Mundus


class Resource:
    def __init__(
        self,
        capacity: int | float | Callable,
        initial_amount: int | float | Callable | None = None,
        label: str | None = None,
    ) -> None:
        """Resource is a class that represents a resource with capacity and amount, any object can use this resource by calling distribute() and return by calling collect() methods.

        Args:
            capacity (int | float | Callable): the capacity of the resource.
            initial_amount (int | float | Callable | None, optional): the initial amount of the resource. Defaults to capacity.
            label (str | None, optional): short description of the resource. Defaults to None.

        Raises:
            ValueError: raise if initial amount is greater than capacity.
        """
        if callable(capacity):
            self._capacity = capacity()
        else:
            self._capacity = capacity
        if initial_amount:
            if callable(initial_amount):
                if initial_amount() > self.capacity:
                    raise ValueError("Initial amount is greater than capacity.")
                else:
                    self._amount = initial_amount()
            else:
                if initial_amount > self.capacity:
                    raise ValueError("Initial amount is greater than capacity.")
                else:
                    self._amount = initial_amount
        else:
            self._amount = self.capacity
        self._label = label
        self._user_records: List[Tuple[object, int | float]] = list() # (user, amount) tracking the usage of individual user
        self._usage_records: List[Tuple[int | float, int | float]] = list() # (time, amount) tracking the usage of the resource over time

    def get(self, amount: int | float) -> None:
        """Get the amount of resource from the resource.

        Args:
            amount (int | float): the amount of resource to get.

        Raises:
            ValueError: raise if amount is greater than the current available amount of resource.
        """        
        if amount > self.amount:
            raise ValueError(f"Not enough amount in Resource {self.label}.")
        else:
            self._amount -= amount
            self.usage_records.append((Mundus.now, self.amount))

    def put(self, amount: int | float) -> None:
        """Put the amount of resource back to the resource.

        Args:
            amount (int | float): the amount of resource to put.

        Raises:
            ValueError: raise if amount is greater than the current used amount of resource.
        """        
        if amount > self.occupied:
            raise ValueError(f"Not enough capacity in Resource {self.label}.")
        else:
            self._amount += amount
            self.usage_records.append((Mundus.now, self.amount))

    def distribute(self, user: object, amount: int | float) -> None:
        """Distribute the amount of resource to the user.

        Args:
            user (object): the user of the resource.
            amount (int | float): the amount of resource to distribute.

        Raises:
            ValueError: raise if amount is greater than the current available amount of resource.
        """        
        if amount > self.amount:
            raise ValueError(f"Not enough amount in Resource {self.label}.")
        else:
            self._amount -= amount
            if user in self.users:
                for index, record in enumerate(self.user_records):
                    if record[0] is user:
                        self.user_records[index] = (user, record[1] + amount)
                        break
            else:
                self.user_records.append((user, amount))
            self.usage_records.append((Mundus.now, self.amount))
            logger.debug(f"Resource {self.label} distributed {amount} to {user}.")

    def collect(self, user: object, amount: int | float | None = None) -> None:
        """Collect the amount of resource from the user.

        Args:
            user (object): the user of the resource.
            amount (int | float | None, optional): the amount of resource to collect. Defaults to None means collecting all resource used by the user.

        Raises:
            ValueError: raise if user is not using the resource or amount is greater than the current used amount of resource.
        """        
        if user not in self.users:
            raise ValueError(f"User {user} is not using Resource {self.label}.")
        if amount is None:
            for index, record in enumerate(self.user_records):
                if record[0] is user:
                    self._amount += record[1]
                    self.usage_records.append((Mundus.now, self.amount))
                    logger.debug(
                        f"Resource {self.label} collected {record[1]} from {user}."
                    )
                    break
        else:
            for index, record in enumerate(self.user_records):
                if record[0] is user:
                    if record[1] < amount:
                        raise ValueError(
                            f"{user} occupied {record[1]} of Resource {self.label}, less than {amount}."
                        )
                    else:
                        self.user_records[index] = (user, record[1] - amount)
                        if self.user_records[index][1] == 0:
                            self.user_records.pop(index)
                        self._amount += amount
                        self.usage_records.append((Mundus.now, self.amount))
                        logger.debug(
                            f"Resource {self.label} collected {amount} from {user}."
                        )
                        break

    def usage(self, duration: int | float | Callable | None = None):
        """Return the usage of the resource in the duration.

        Args:
            duration (int | float | Callable | None, optional): the duration to trace back in time. Defaults to None.
        """        
        if duration:
            if callable(duration):
                after = Mundus.now - duration()
            else:
                after = Mundus.now - duration
            if after < 0:
                after = 0
            usage_records = [
                usage_record
                for usage_record in self.usage_records
                if usage_record[0] >= after
            ]
            if len(usage_records) == 0:
                return 1 - (self.amount / self.capacity)
            else:
                if usage_records[-1][0] - after == 0:
                    return 1 - (usage_records[-1][1] / self.capacity)
                else:
                    weighted_overall_amount = 0
                    for index, record in enumerate(usage_records):
                        if index == 0:
                            weighted_overall_amount += record[1] * (record[0] - after)
                        elif index == len(usage_records) - 1:
                            weighted_overall_amount += record[1] * (
                                Mundus.now - usage_records[index - 1][0]
                            )
                        else:
                            weighted_overall_amount += record[1] * (
                                record[0] - usage_records[index - 1][0]
                            )
                    return 1 - (
                        weighted_overall_amount / (usage_records[-1][0] - after)
                    )
        else:
            return 1 - (self.amount / self.capacity)

    @property
    def amount(self) -> int | float:
        """Return the current available amount of the resource."""
        return self._amount

    @property
    def capacity(self) -> int | float:
        """Return the capacity of the resource."""
        return self._capacity

    @property
    def occupied(self) -> int | float:
        """Return the current occupied amount of the resource."""
        return self.capacity - self.amount

    @property
    def label(self) -> str | None:
        """Return the label of the resource."""
        return self._label

    @property
    def user_records(self) -> List[Tuple[object, int | float]]:
        """Return the usage records of the resource per user."""
        return self._user_records

    @property
    def usage_records(self) -> List[Tuple[int | float, int | float]]:
        """Return the usage records of the resource, for calculating the usage of the resource in a duration."""
        return self._usage_records

    @property
    def users(self) -> List[object]:
        """Return the users of the resource."""
        return [record[0] for record in self._user_records]
